#include <math.h>
#include "kask.h"
#include "kasktri.h"
#include "kaskass.h"
#include "kasksol.h"

/*  ======================================================================

    BPX -- Realisierung entsprechend :

    Folkmar A. Bornemann: An Adaptive Multilevel Approach to Parabolic 
                          Equations in Two Space Dimensions,
		          Technical Report TR 91-7, Konrad-Zuse-Zentrum,
			  Berlin (1991)	

    Bemerkungen:
    ------------

       1.  Die Felder S^+, U^+, l sind in Sparsematrix-Technik realisiert,
           dies macht die Verwendung des "indexPlusArray"'s noetig.

       2.  Der Zugriff auf die Punkte in M_k erfolgt indirekt ueber die
           verfeinerte Kante der Tiefe k-1. Dies stellt die Beziehung
           Mittelpunkt <-> Eckpunkt her.

       3.  "MKMinus" bezeichnet das (mengentheoretische) Komplement von
           M_k^+ ("MKPlus") in M_(k+1). Dies entspricht den wiederholten
           Fallunterscheidungen.

     ====================================================================== */										

/*   *************
1.0  Deklarationen
     ************* */

/*   ---------------------
1.1  Externe Deklarationen
     --------------------- */

INT             volumeKnown;

extern	INT     ArrayAlloc();
extern 	void    ArrayFree();

/*   -----------------------
1.2  Statische Deklarationen
     ----------------------- */

static REAL     *sPlus = nil, *uPlus = nil, *bpxData = nil;
static INT      *lArray = nil, *indexPlusArray = nil;
static char     *dataNames[] = {"qBar", "volumeWeight"};

static REAL     meanInitVolume, qData, q1;
static INT      offset[MAX_DEPTH], offsetPlus, counter, nMinus, nPlus;
static INT      depth, maxDepth, BY, BX, psiNum;

static INT      AddInitVolume(), InitVolume(), UnmarkPoints();
static void     InitZeroUPlus();
static REAL     V(), ValV();

/*   **********************************
2.0  BPX - Initialisierung fuer KASKADE
     ********************************** */
void 
InitBpx()
{
   if (bpxData == nil)
     {
       bpxData =
          (REAL *) NewParamList("bpxData", 2, sizeof(REAL), 1, dataNames, 
                               nil, nil);

       bpxData[0] = 10.0;
       bpxData[1] = 6.0;
     }
   return;
}

/*   *********************************************
3.0  Sparse Matrix Routinen fuer benoetigte Felder
     ********************************************* */

static INT 	
ArrayAlloc()
{
	int		i;

	psiNum = 5*((actTriang->noOfPoints) - (actTriang)->noOfInitPoints)/3 + (actTriang)->noOfInitPoints;

	sPlus = (REAL*) ZIBAlloc((long) sizeof(REAL)*psiNum);
	if (sPlus == nil)
		return false;

	uPlus = (REAL*) ZIBAlloc((long) sizeof(REAL)*psiNum);
	if (uPlus == nil)
		return false;

	lArray = (int*) ZIBAlloc((long) sizeof(int)*psiNum);
	if (lArray == nil)
		return false;

	indexPlusArray = (int*) ZIBAlloc((long) sizeof(int)*psiNum);
	if (indexPlusArray == nil)
		return false;

	for (i = 0; i < psiNum; i++) {
		sPlus[i] = ZERO;
		uPlus[i] = ZERO;
		lArray[i] = 0;
		indexPlusArray[i] = 0;
	}

	return true;
}

static void
ArrayFree()
{
	if (sPlus == nil)
		return;

	ZIBFree((PTR) sPlus);
	ZIBFree((PTR) uPlus);
	ZIBFree((PTR) lArray);
	ZIBFree((PTR) indexPlusArray);

	return;
}

/*   ***************
4.0  Initialisierung
     *************** */

static int
InitArrays()
{
	if(!volumeKnown) {
		ArrayFree();
		if (!ArrayAlloc()) {
			sprintf(globBuf,"Not enough memory (ArrayAlloc)\n");
			ZIBStdOut(globBuf);
			return false;
		}
		meanInitVolume = ZERO;
		ApplyT3(AddInitVolume, initial);
		meanInitVolume /= (actTriang->noOfInitTetrahedra);
		qData = bpxData[0] * meanInitVolume;

		ApplyT3(InitVolume, all);
	}
	InitZeroUPlus();

	return true;
}

/*   ------------------------------
4.1  Hilfsroutinen: Initialisierung
     ------------------------------ */

static void
InitZeroUPlus()
{
	int		i;

	for (i = 0; i < psiNum; i++) 
		uPlus[i] = ZERO;
	return;
}

static int
InitPoint(p)
	PT		*p;
{
	uPlus[p->index] = RA(p, BX);
	p->l = maxDepth;
	p->counter = p->index;

	return true;
}

static int
AddInitVolume(t)
	TR3		*t;
{
	meanInitVolume += t->volume;
	return true;
}

static INT 
InitVolume(t)
	TR3             *t;
{
	PT	      **pts = &(t->p1);
	TR3	       *tInit;
	REAL            volume;
        REAL fVals[NO_OF_F_VALS], xm, ym, zm;
	int		i;

	tInit = t;
        while ((tInit->father) != nil)
                tInit = tInit->father;

	volume = (t->volume) * bpxData[1] / (tInit->volume);

        xm =t->p1->x; 
        ym =t->p1->y; 
        zm =t->p1->z; 


        (actProblem->F)(xm, ym, zm, t->class,fVals,-1);

        volume *= fVals[0];
	for (i = 0; i < 4; i++)
		sPlus[pts[i]->index] += volume;

	return true;
}

/*   ************************
5.0  Routinen fuer Tiefe Null
     ************************ */

static int
ResAndLNaught(p)
	PT		*p;
{
	int		index;

	counter++;

	index = offset[0]+counter;
	lArray[index] = p->l;
	indexPlusArray[index] = p->counter;
	RA(p, BY) = uPlus[p->counter];
	
	return true;
}

static int
ValueOnNaught(p)
	PT		*p;
{
	RA(p, BY) = q1 * RA(p, BY) + V(p, 0);

	return true;
}

/*   *******************
6.0  Routinen auf MKPlus
     ******************* */

static int
MarkPointMKPlus(p)
	PT		*p;
{
	if ( (!(p->mark)) && ((p->boundP)!=DIRICHLET) ) {
		nPlus++;
		p->mark = -nPlus;
	}

	return true;
}

static int
RestrictMKPlus(p)
	PT		*p;
{
	int		index, indexPlus;

	if ((p->mark)>=0)
		return true;

	indexPlus = offsetPlus - p->mark;
        p->mark = nMinus - p->mark;
	index = offset[depth] + p->mark;
	lArray[index] = p->l;
	indexPlusArray[index] = p->counter;
	p->l = depth - 1;
	p->counter = indexPlus;

	uPlus[indexPlus] += uPlus[indexPlusArray[index]];

	if(!volumeKnown)
		sPlus[indexPlus] += sPlus[indexPlusArray[index]];

	return true;
}

static int
AddMKPlus(p)
	PT		*p;
{
	if ((p->mark) >=0)
		return true;
	p->mark = nMinus - p->mark;
	RA(p, BY) += V(p, depth);

	return true;
}

/*   ********************
7.0  Routinen auf MKMinus
     ******************** */

static int
MarkPointMKMinus(p)
	PT		*p;
{
	if ( (!(p->mark)) && ((p->boundP)!=DIRICHLET) ){
		nMinus++;
		p->mark = nMinus;
	}

	return true;
}

static INT 
RestrictMKMinus(p,ed)
	PT		*p;
	EDG		*ed;
{
	PT		*p1 = ed->p1, *p2 = ed->p2;
	REAL		uVal, sVal;
	int		index, indexPlus;

	if ( ((p->mark)<=0) || ((p->mark) > nMinus) )
		return true;

	index = offset[depth]+(p->mark);
	lArray[index] = p->l;
	indexPlusArray[index] = p->counter;

	uVal = uPlus[p->counter];

	if (p1->mark) {
		index = offsetPlus + p1->mark - nMinus;
		uPlus[index] += HALF * uVal;
	}

	if (p2->mark) {
		index = offsetPlus + p2->mark - nMinus;
		uPlus[index] += HALF * uVal;
	}

	if (!volumeKnown) {
		sVal = sPlus[p->counter];

		if (p1->mark) {
			index = offsetPlus + p1->mark - nMinus;
			sPlus[index] += HALF * sVal;
		}

		if (p2->mark) {
			index = offsetPlus + p2->mark - nMinus;
			sPlus[index] += HALF * sVal;
		}
	}

	return true;
}

static int
AddMKMinus(p)
	PT		*p;
{
	if ( ((p->mark)<=0) || ((p->mark)>nMinus) )
		return true;

	RA(p, BY) += V(p, depth);

	return true;
}

/*   **************
8.0  Die Funktion V
     ************** */

static REAL
V(p,k)
	PT		*p;
	int		k;
{
	int		index;

	if (k == 0) {
		counter++;
		index = offset[0] + counter;
	} else
		index = offset[k] + p->mark;

	return ValV(indexPlusArray[index], k, lArray[index]);
}

/*   --------------------------------------------------
8.1  Die eigentliche Berechnung von V, das Herz von BPX
     -------------------------------------------------- */

static REAL
ValV(index, k0, k1)
	int		index, k0, k1;
{
	REAL		val;

	val = 1.0 / ( qData + (pow(4.0, (double) (k0))) );		
	if (k1 < maxDepth)
		val -= 1.0 / ( qData + (pow(4.0, (double) (k1+1))) );	
	
	val *= uPlus[index] / sPlus[index];
	
	return val;
}

/*   ******************
9.0  Die Kantenroutinen
     ****************** */

static int
MarkPointsOnEdge(ed)
	EDG		*ed;
{
	MarkPointMKPlus(ed->p1);
	MarkPointMKPlus(ed->p2);
	MarkPointMKMinus(PM(ed));

	return true;
}

static int
RestrictOnEdge(ed)
	EDG		*ed;
{
	RestrictMKPlus(ed->p1);
	RestrictMKPlus(ed->p2);
	RestrictMKMinus(PM(ed),ed);

	return true;
}
	
static int
AddOnEdges(ed)
	EDG		*ed;
{
	AddMKPlus(ed->p1);
	AddMKPlus(ed->p2);
	AddMKMinus(PM(ed));

	return true;
}

static INT 
Interpolate(ed)
	EDG		*ed;
{
	RA(PM(ed), BY) = HALF * (RA(ed->p1, BY) + RA(ed->p2, BY));
	
	return true;
}

/*    *************************
10.0  Restriktion und Summation
      ************************* */

static void
Restrictions()
{
	InitArrays();
	ApplyP(InitPoint, allNotDirichlet);
	offset[maxDepth] = - 1;
	offsetPlus = actTriang->noOfPoints - 1;

	for (depth = maxDepth; depth > 0; depth--) {
		nPlus = 0;
		nMinus = 0;
		ApplyEDepth(MarkPointsOnEdge, depth - 1);
		ApplyEDepth(RestrictOnEdge, depth - 1);
		ApplyEDepth(UnmarkPoints, depth - 1);

		offset[depth - 1] = offset[depth] + nMinus + nPlus;
		offsetPlus = offsetPlus + nPlus;
	}

	counter = 0;
	ApplyP(ResAndLNaught, initialNotDirichlet);

	return;
}

static void
SumUpBPXTerms()
{
	q1 = 1.0 / (qData + 1.0);

	counter = 0;
	ApplyP(ValueOnNaught, initialNotDirichlet);
	for (depth = 1; depth <= maxDepth; depth++) {
		ApplyEDepth(Interpolate, depth - 1);
		nPlus = 0;
		nMinus = 0;
		ApplyEDepth(MarkPointsOnEdge, depth - 1);
		ApplyEDepth(AddOnEdges, depth - 1);
		ApplyEDepth(UnmarkPoints, depth - 1);
	}

	return;
}

/*    ****************
11.0  Die Hauptroutine
      **************** */

				
void
bpxPcxMul(x, y)
	int		x,y;
{
	REAL		beta;

	BX = x;
	BY = y;

	SetZeroField(BY);
	
	maxDepth = actTriang->maxDepth;
	Restrictions();
	Lev0DirectSol(BY);
	SumUpBPXTerms();

	return;
}

/*    ************
12.0  Aux--Routine
      ************ */

static int
UnmarkPoints(ed)
	EDG		*ed;
{
	PT		**pts = &(ed->p1);
	int		i;

	for (i = 0; i < 3; i++)
		pts[i]->mark = false;

	return true;
}

	



	
